@@ -27,6 +27,8 @@
 #include <sys/socket.h>
 #include <string.h>
 #include <unistd.h>
+#include <errno.h>
+#include <dirent.h>
 #include <getopt.h>
 #include <stdarg.h>
 #include <assert.h>
@@ -38,6 +40,44 @@
 #include "qemuarch.h"
 #include "file.h"
 
+#define STR(x) #x
+#define XSTR(x) STR(x)
+
+#define PBUILDER_INIT_VERSION 1
+#define BUILDDIR "/build"
+#define CHROOT_HOOKDIR BUILDDIR"/hooks"
+
+#define EXECUTE_HOOKS_INDENT(indent, prefix) \
+indent"if [ -d \""CHROOT_HOOKDIR"\" ]; then\n" \
+indent"    for fn in \""CHROOT_HOOKDIR"/"prefix"\"[0-9][0-9]* ; do\n" \
+indent"        case \"$fn\" in\n" \
+indent"            \""CHROOT_HOOKDIR"/"prefix"\"'[0-9][0-9]*')\n" \
+indent"                log.d \"no hooks of type "prefix" found -- ignoring\"\n" \
+indent"                ;;\n" \
+indent"            *~)\n" \
+indent"                log.w \"skipping an editor backup file $fn\"\n" \
+indent"                ;;\n" \
+indent"            *)\n" \
+indent"                if [ -x \"$fn\" ]; then\n" \
+indent"                    log.i \"user script $fn starting\"\n" \
+indent"                    \""CHROOT_HOOKDIR"/$(basename \"$fn\")\"\n" \
+indent"                    log.i \"user script $fn finished\"\n" \
+indent"                else\n" \
+indent"                    if [ -f \"$fn\" ]; then\n" \
+indent"                        filetype=$(basename \"$fn\")\n" \
+indent"                        log.w \"execute priv not set on file $filetype, not executing.\"\n" \
+indent"                    else\n" \
+indent"                        # Should it reach here ? This case should be caught in the above case.\n" \
+indent"                        log.w \"no hooks of type ${prefix} found -- internal error in logic\"\n" \
+indent"                    fi\n" \
+indent"                fi\n" \
+indent"                ;;\n" \
+indent"        esac\n" \
+indent"    done\n" \
+indent"fi\n"
+
+#define EXECUTE_HOOKS(prefix) EXECUTE_HOOKS_INDENT("", prefix)
+
 /*
  * example exit codes:
  *
@@ -128,6 +168,23 @@ static FILE* create_script(const char* mountpoint, const char* relative_path)
   FILE *ret = NULL;
 
   asprintf(&s, "%s/%s", mountpoint, relative_path);
+
+  /* Create directories in relative_path */
+  for (char *c = s + strlen(mountpoint) + 1; *c; ++c)
+    {
+      if (*c != '/') continue;
+      *c = 0;
+      if (mkdir(s, 0777) && errno != EEXIST)
+	{
+	  log_printf(log_error,
+              "Could not create directory '%s': %s",
+              s,
+              strerror(errno));
+	  goto fail;
+	}
+      *c = '/';
+    }
+
   if(!(f=fopen(s, "w")))
     goto fail;
   if(chmod(s, 0700))
@@ -141,43 +198,211 @@ static FILE* create_script(const char* mountpoint, const char* relative_path)
   return ret;
 }
 
-/* minimally fix terminal I/O */
-static void fix_terminal(void)
-{
-  struct termios t;
+static struct termios saved_termios;
 
+static void save_termios(void)
+{
   if (isatty(1))
     {
-      tcgetattr(1, &t);
-      t.c_lflag |= ECHO;
-      tcsetattr(1, TCSANOW, &t);
+      tcgetattr(1, &saved_termios);
     }
 }
 
+static void restore_termios(void)
+{
+  if (isatty(1))
+    {
+      tcsetattr(1, TCSANOW, &saved_termios);
+    }
+}
+
+static int copy_file_contents_to_temp(const char* orig,
+				      const char* tempdir,
+				      const char* tempname)
+{
+  char* temppath;
+  int ret;
+  if (tempname == NULL)
+    tempname = basename(orig);
+
+  asprintf(&temppath, "%s/%s", tempdir, tempname);
+  ret=copy_file(orig, temppath);
+  if (ret == -1)
+    {
+      log_printf(log_error, "Copy file error in %s to %s",
+          orig, temppath);
+      goto out;
+    }
+out:
+  free(temppath);
+  return ret;
+}
+
+static int copy_file_contents_in_temp(FILE *f,
+				      const char* tempname,
+				      const char* targetdir,
+				      const char* targetname)
+{
+  int trailing_slash = targetdir[strlen(targetdir)-1] == '/';
+  const char* sep = trailing_slash ? "" : "/";
+  fprintf(f,
+	  "log.i \"copying %s%s%s from temporary location\"\n"
+	  "mkdir -p %s\n"
+	  "cp $BUILDDIR/%s %s%s%s || log.e \"Copy failed\"\n",
+	  targetdir, sep, targetname,
+	  targetdir,
+	  tempname,
+	  targetdir, sep, targetname);
+  return 0;
+}
+
 static int copy_file_contents_through_temp(FILE* f,
 					   const char* orig,
-					   const char* temppath,
+					   const char* tempdir,
 					   const char* targetdir)
 {
-  char* tempname;
   int ret;
   char* file_basename = basename(orig);
-  asprintf(&tempname, "%s/%s", temppath, file_basename);
-  ret=copy_file(orig, tempname);
-  if (ret == -1)
+  char* tempname = NULL;
+  char* tempinput = NULL;
+  if (0>asprintf(&tempinput,
+		 "%s/input",
+		 tempdir))
     {
-      fprintf(f, "E: Copy file error in %s to %s\n",
-	      orig, tempname);
+      log_printf(log_error,
+	  "failed to allocate string for '%s/input': %s",
+	  tempdir,
+	  strerror(errno));
+      ret = 1;
+      goto out;
     }
 
-  free(tempname);
-
-  fprintf(f,
-	  "echo \"I: copying %s/%s from temporary location\"\n"
-	  "cp $PBUILDER_MOUNTPOINT/%s %s/%s || echo \"E: Copy failed\"\n",
-	  targetdir, file_basename,
+  if (0>asprintf(&tempname,
+		 "input/%s",
+		 file_basename))
+    {
+      log_printf(log_error,
+	  "failed to allocate string for 'input/%s': %s",
 	  file_basename,
-	  targetdir, file_basename);
+	  strerror(errno));
+      ret = 1;
+      goto out;
+    }
+
+  if (mkdir(tempinput, 0777) && errno != EEXIST)
+    {
+      log_printf(log_error,
+	  "failed to create directory '%s': %s",
+	  tempinput,
+	  strerror(errno));
+      ret = 1;
+      goto out;
+    }
+
+  ret = copy_file_contents_to_temp(orig, tempinput, file_basename);
+  if (ret != 0)
+    goto out;
+
+  ret = copy_file_contents_in_temp(f, tempname, targetdir, file_basename);
+
+out:
+  if (tempname != NULL)
+    {
+      free(tempname);
+    }
+  if (tempinput != NULL)
+    {
+      free(tempinput);
+    }
+  return ret;
+}
+
+static int copy_hookdir(const char *hookdir, const char *tmp)
+{
+  int ret = 0;
+  struct dirent *dirp;
+  DIR *hd = opendir(hookdir);
+  char *hookstmp = NULL;
+  if (0>asprintf(&hookstmp,
+		 "%s/hooks",
+		 tmp))
+    {
+      log_printf(log_error,
+	  "Error allocating string for '%s/hooks': %s",
+	  tmp,
+	  strerror(errno));
+      ret = 1;
+      goto out;
+    }
+
+  if (mkdir(hookstmp,0777))
+    {
+      log_printf(log_error,
+	  "Error creating directory for hooks: %s",
+	  strerror(errno));
+      ret = 1;
+      goto out;
+    }
+
+  if (hd == NULL)
+    {
+      log_printf(log_error,
+	  "Error copying hooks from '%s': %s",
+	  hookdir,
+	  strerror(errno));
+      ret = 1;
+      goto out;
+    }
+
+  while ((dirp = readdir(hd)) != NULL)
+    {
+      struct stat st;
+      char *src = NULL;
+
+      if (0>asprintf(&src,
+		     "%s/%s",
+		     hookdir,
+		     dirp->d_name))
+	{
+	  log_printf(log_error,
+	      "Error allocating string for '%s/%s': %s",
+	      hookdir,
+	      dirp->d_name,
+	      strerror(errno));
+	  ret = 1;
+	  goto out;
+	}
+      if (0>stat(src, &st))
+	{
+	  log_printf(log_error,
+	      "Error calling stat '%s': %s",
+	      src,
+	      strerror(errno));
+	  free(src);
+	  ret = 1;
+	  goto out;
+	}
+
+      if ((st.st_mode & S_IFMT) != S_IFREG)
+	{
+	  // Not a regular file
+	  free(src);
+	  continue;
+	}
+
+      copy_file_contents_to_temp(src, hookstmp, basename(src));
+      free(src);
+    }
+
+out:
+  if (hookstmp != NULL)
+    {
+      free(hookstmp);
+    }
+  if (hd != NULL)
+    {
+      closedir(hd);
+    }
   return ret;
 }
 
@@ -224,6 +449,8 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       return -1;
     }
 
+  save_termios();
+
   fflush(NULL);
   if ((child=fork()))
     {
@@ -239,7 +466,7 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	  if (0<waitpid(child, &status, WNOHANG))
 	    {
 	      /* child has exited */
-	      printf("\n   -> qemu exited unexpectedly: %d\n", status);
+	      log_printf(log_error, "qemu exited unexpectedly: %d", status);
 	      break;
 	    }
 	  FD_SET(sp[0],&readfds);
@@ -261,19 +488,19 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 				       qemu_keyword, strlen(qemu_keyword)))!=0)
 		    {
 		      exit_code = atoi(matchptr + strlen(qemu_keyword));
-		      printf("\n  -> received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)\n",
-			     exit_code,
-			     child);
+		      log_printf(log_info, "received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)",
+			  exit_code,
+			  child);
 
 		      assert(child != 0);assert(child > 0);
 
 		      if (!kill(child, SIGTERM))
-			printf("   -> successfully killed qemu\n");
+			log_printf(log_info, "successfully killed qemu");
 		      else
-			perror("   -> failed to kill qemu? :");
+			log_perror("failed to kill qemu?");
 		      if (-1==waitpid(child, &status, 0))
 			{
-			  perror("qemubuilder: waitpid");
+			  log_perror("qemubuilder: waitpid");
 			}
 		      break;
 		    }
@@ -282,7 +509,7 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 	    }
 	  else
 	    {
-	      perror("select");
+	      log_perror("select");
 	      break;
 	    }
 	}
@@ -304,14 +531,24 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       const int MAX_ARGS = 40;
       char *argv[MAX_ARGS];
       int i;
+      int is_tty = isatty(1);
+      const char *term;
+      if (is_tty)
+	{
+	  term = getenv("TERM");
+	  if (!term)
+	    term = ""; /* Use whatever getty defaults to */
+	}
+      else
+	term = "dumb";
 
       if (qemu == NULL || machine == NULL) {
-	fprintf(stderr, "Your architecture %s does not seem to be supported\n", pc->arch);
+	log_printf(log_error, "Your architecture %s does not seem to be supported", pc->arch);
 	exit(1);
       }
 
       if (kernel_image == NULL || !strcmp(kernel_image, "")) {
-	fprintf(stderr, "No KERNEL_IMAGE defined in pbuilderrc\n");
+	log_printf(log_error, "No KERNEL_IMAGE defined in pbuilderrc");
 	exit(1);
       }
 
@@ -319,13 +556,13 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 
       if (hda_format == NULL)
 	{
-	  fprintf(stderr, "Unknown format for disk image %s\n", hda);
+	  log_printf(log_error, "Unknown format for disk image %s", hda);
 	  exit(1);
 	}
 
       if (hdb_format == NULL)
 	{
-	  fprintf(stderr, "Unknown format for disk image %s\n", hdb);
+	  log_printf(log_error, "Unknown format for disk image %s", hdb);
 	  exit(1);
 	}
 
@@ -340,9 +577,10 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       /* panic < 0 means reboot immediately on panic; this will actually
          halt as -no-reboot is given to qemu */
       asprintf(&append_command,
-	       "root=/dev/%sa quiet init=/pbuilder-run console=%s panic=-1",
+	       "root=/dev/%sa quiet init=/usr/bin/setsid console=%s panic=-1 -- -c -w /sbin/getty -n -l /pbuilder-run -8 -L - %s",
 	       qemu_arch_diskdevice(pc),
-	       qemu_arch_tty(pc->arch));
+	       qemu_arch_tty(pc->arch),
+	       term);
 
       dup2(sp[1],1);
       dup2(sp[1],2);
@@ -374,7 +612,14 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       argv[argc++]="-append";
       argv[argc++]=append_command;
       argv[argc++]="-serial";
-      argv[argc++]="stdio";
+      if (is_tty)
+	{
+	  argv[argc++]="mon:stdio";
+	}
+      else
+	{
+	  argv[argc++]="stdio";
+	}
       argv[argc++]="-net";
       argv[argc++]="user";
       if (!strcmp(machine, "virt")) {
@@ -395,24 +640,28 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
       argv[argc]=NULL;
       assert(argc < MAX_ARGS);
 
-      printf("  forking qemu: ");
+      log_begin(log_info);
+      log_middle(log_info, "forking qemu:");
       for (i=0; i<argc; ++i)
 	{
-	  printf("%s ", argv[i]);
+	  if (strchr(argv[i], ' '))
+	    log_middle(log_info, " '%s'", argv[i]);
+	  else
+	    log_middle(log_info, " %s", argv[i]);
 	}
-      printf("\n");
+      log_end(log_info);
 
       execvp(argv[0], argv);
-      perror("fork_qemu");
+      log_perror("fork_qemu");
       exit (1);
     }
   else
     {
-      perror("fork");
+      log_perror("fork");
       return -1;
     }
 
-  fix_terminal();
+  restore_termios();
   return exit_code;
 }
 
@@ -446,6 +695,140 @@ static char* get_current_time_string(void)
   return timestring;
 }
 
+static void write_first_stage(FILE *f, const struct pbuilderconfig* pc)
+{
+  fprintf(f,
+	  "#!/bin/bash\n"
+	  "echo \n"
+	  /* Can't use log.i or check LOGLEVEL in first stage, since this lives
+	   * inside the base chroot, and has not yet mounted the input disk. */
+	  "echo 'I: qemu-pbuilder first-stage' \n"
+	  "export PBUILDER_INIT_VERSION="XSTR(PBUILDER_INIT_VERSION)"\n"
+	  "export PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'\n"
+	  "stty sane\n"
+	  "[ -d /proc/1 ] || mount -n /proc /proc -t proc\n"
+	  "ln -s /dev/shm /run/shm\n"
+	  "mkdir /run/lock\n"
+	  "chmod 01777 /run/lock\n"
+	  "mount -t tmpfs tmpfs /run/lock\n"
+	  "mount -n -o rw,remount / || mount -v -n -o rw,remount /dev/root \n"
+	  "find /tmp -mindepth 1 -maxdepth 1 -print0 | xargs -0 rm -rf --\n"
+	  "mount -t tmpfs tmpfs /tmp\n"
+	  "mkdir /dev/shm\n"
+	  "chmod 01777 /dev/shm\n"
+	  "mount -t tmpfs tmpfs /dev/shm\n"
+	  "ln -fs /proc/mounts /etc/mtab\n"
+	  "export BUILDDIR="BUILDDIR"\n"
+	  "mkdir -p $BUILDDIR\n"
+	  "ln -s $BUILDDIR /tmp/buildd\n"
+	  "mount -n -t ext3 /dev/%sb $BUILDDIR \n"
+	  "$BUILDDIR/input/pbuilder-run \n",
+	  qemu_arch_diskdevice(pc)
+	  );
+}
+
+static void write_second_stage_header(FILE *f, int debug_shell) {
+  const char *log_level;
+
+  switch (log_get_filter_level()) {
+    case log_debug: log_level = "D"; break;
+    case log_info:  log_level = "I"; break;
+    case log_warn:  log_level = "W"; break;
+    case log_error: log_level = "E"; break;
+    default:        log_level = "I"; break;
+  }
+
+  fprintf(f,
+	  "#!/bin/bash\n"
+
+	  /* Will not be auto (using log_get_use_colors, not _unresolved) */
+	  "export USECOLORS=\"%s\"\n"
+	  "export LOGLEVEL=\"%s\"\n"
+
+	  /* Bash logging taken from pbuilder */
+	  "# Log a message\n"
+	  "# message is of a format\n"
+	  "#  E: error message\n"
+	  "#  W: warning message\n"
+	  "#  I: informational message\n"
+	  "_log() {\n"
+	  "    set -u\n"
+	  "    local color=\"$1\" ; shift\n"
+	  "    local red='\\033[0;31m'\n"
+	  "    local yellow='\\033[1;33m'\n"
+	  "    local blue='\\033[0;34m'\n"
+	  "    local reset='\\033[0m'\n"
+	  "    case \"$USECOLORS\" in\n"
+	  "        yes)\n"
+	  "            printf \"${!color}${*}${reset}\\n\"\n"
+	  "            ;;\n"
+	  "        no)\n"
+	  "            printf \"${*}\\n\"\n"
+	  "            ;;\n"
+	  "        *)\n"
+	  "            printf \"malformed value of USECOLORS: [%%s]\\n\" \"$USECOLORS\" >&2\n"
+	  "            exit 1\n"
+	  "            ;;\n"
+	  "    esac\n"
+	  "    set +u\n"
+	  "}\n"
+	  "function log() {\n"
+	  "    case \"$*\" in\n"
+	  "        \"E: \"*)\n"
+	  "            _log 'red' \"$*\" >&2\n"
+	  "            ;;\n"
+	  "        \"W: \"*)\n"
+	  "            _log 'yellow' \"$*\" >&2\n"
+	  "            ;;\n"
+	  "        \"I: \"*)\n"
+	  "            _log 'reset' \"$*\"\n"
+	  "            ;;\n"
+	  "        \"D: \"*)\n"
+	  "            _log 'blue' \"$*\"\n"
+	  "            ;;\n"
+	  "        *)\n"
+	  "            echo \"malformed log message: $*\" >&2\n"
+	  "            exit 1\n"
+	  "            ;;\n"
+	  "    esac\n"
+	  "}\n"
+
+	  "log.e() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I|W|E) log \"E: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.w() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I|W) log \"W: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.i() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D|I) log \"I: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+	  "log.d() {\n"
+	  "    case \"$LOGLEVEL\" in\n"
+	  "        D) log \"D: $*\" ;;\n"
+	  "    esac\n"
+	  "}\n"
+
+	  /* define function to terminate qemu */
+	  "function exit_from_qemu() {\n"
+	  "    %s\n"
+	  "    sync\n"
+	  "    sync\n"
+	  "    sleep 1s\n"		/* sleep before sending dying message */
+	  "    log.i 'qemu-pbuilder %s'\"$1\"\n"
+	  "    sleep 1s\n"
+	  "    halt -f -p\n"		/* just halt myself if possible */
+	  "}\n",
+	  log_get_use_colors() == log_use_colors_yes ? "yes" : "no",
+	  log_level,
+	  debug_shell?"log.i \"Debug shell\"; /bin/bash":"",
+	  qemu_keyword);
+}
 
 /**
  * Invoke qemu, and run the second-stage script within QEMU.
@@ -475,7 +858,7 @@ static int run_second_stage_script
   if (mkdir(pc->buildplace,0777))
     {
       /* could not create the buildplace here. */
-      perror("mkdir");
+      log_perror("mkdir");
       goto out;
     }
 
@@ -487,64 +870,86 @@ static int run_second_stage_script
   ret=create_ext3_block_device(workblockdevicepath, 1);
   loop_mount(workblockdevicepath, pc->buildplace);
 
-  f = create_script(pc->buildplace, "pbuilder-run");
-  fprintf(f,
-	  "#!/bin/bash\n"
-
-	  /* define function to terminate qemu */
-	  "function exit_from_qemu() {\n"
-	  "sync\n"
-	  "sync\n"
-	  "sleep 1s\n"		/* sleep before sending dying message */
-	  "echo ' -> qemu-pbuilder %s$1'\n"
-	  "sleep 1s\n"
-	  "halt -f -p\n"	/* just halt myself if possible */
-	  "}\n",
-	  qemu_keyword);
-
+  f = create_script(pc->buildplace, "input/pbuilder-run");
+  write_second_stage_header(f, 0);
   fprintf(f,
 	  /* main code */
 	  "echo \n"
-	  "echo ' -> qemu-pbuilder second-stage' \n"
-	  //TODO: copy hook scripts
+	  "log.i 'qemu-pbuilder second-stage' \n"
+	  // Remove compatibility symlink
+	  "rm \"$BUILDDIR\"/pbuilder-run\n"
 	  //"mount -n /proc /proc -t proc\n" // this is done in first stage.
-	  "echo '  -> setting time to %s' \n"
+	  "if [ \"${PBUILDER_INIT_VERSION:-0}\" -lt "XSTR(PBUILDER_INIT_VERSION)" ]; then\n"
+	  "    log.e \"qemubuilder init script is out of date (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
+	  "    log.e \"Please run qemubuilder --update\"\n"
+	  "    exit_from_qemu 1\n"
+	  "elif [ \"${PBUILDER_INIT_VERSION:-0}\" -gt "XSTR(PBUILDER_INIT_VERSION)" ]; then\n"
+	  "    log.e \"qemubuilder init script is newer than expected (${PBUILDER_INIT_VERSION:-0} < "XSTR(PBUILDER_INIT_VERSION)")\"\n"
+	  "    exit_from_qemu 1\n"
+	  "fi\n"
+	  "log.i 'setting time to %s' \n"
 	  "date --set=\"%s\"\n"
-	  "echo '  -> configuring network' \n"
+	  "log.i 'configuring network' \n"
 	  "ifconfig -a\n"
 	  "export IFNAME=`/sbin/ifconfig -a | grep eth | head -n1 | awk '{print $1}'`\n"
 	  "dhclient $IFNAME\n"
-	  "mkdir -p /tmp/buildd\n"
-	  "$PBUILDER_MOUNTPOINT/run-copyfiles\n"
+	  "mkdir -p \""BUILDDIR"\"\n"
+	  "%s\n"
+	  "$BUILDDIR/input/run-copyfiles\n"
 	  "hostname pbuilder-$(cat /etc/hostname)\n"
-	  //TODO: run G hook
 	  "%s\n"
 	  //TODO: I can mount /var/cache/apt/archives from some scratch space to not need this:
 	  "apt-get clean || true\n"
 	  "exit_from_qemu 0\n",
 	  timestring,
 	  timestring,
+	  pc->hookdir && pc->hookdir[0] ? "mkdir -p \""CHROOT_HOOKDIR"\"" : "",
 	  commandline);
   fclose(f);
+  {
+    char *compat_symlink;
+    if (0>asprintf(&compat_symlink,
+		   "%s/pbuilder-run",
+		   pc->buildplace))
+      {
+	log_printf(log_error,
+	    "Failed to allocate string for compatibility symlink path '%s/pbuilder-run': %s",
+	    pc->buildplace,
+	    strerror(errno));
+	goto out;
+      }
+    if (symlink("input/pbuilder-run", compat_symlink))
+      {
+	log_printf(log_error,
+	    "Failed to create compatibility symlink: %s",
+	    strerror(errno));
+	free(compat_symlink);
+	goto out;
+      }
+    free(compat_symlink);
+  }
 
   /* copy files script */
-  f = create_script(pc->buildplace, "run-copyfiles");
+  f = create_script(pc->buildplace, "input/run-copyfiles");
   copy_file_contents_through_temp(f, "/etc/hosts", pc->buildplace, "/etc");
   copy_file_contents_through_temp(f, "/etc/hostname", pc->buildplace, "/etc");
   /* copy inputfile */
   for (i=0; pc->inputfile[i]; ++i)
     {
-      copy_file_contents_through_temp(f, pc->inputfile[i], pc->buildplace, "/tmp/buildd");
+      copy_file_contents_to_temp(pc->inputfile[i], pc->buildplace, NULL);
+    }
+  if (pc->hookdir != NULL && pc->hookdir[0])
+    {
+      copy_hookdir(pc->hookdir, pc->buildplace);
     }
   fclose(f);
 
   /* do I not need to copy /etc/pbuilderrc, and ~/.pbuilderrc to inside chroot? */
-  /* TODO: hooks probably need copying here. */
   /* TODO: recover aptcache */
 
   if(hostcommand1)
     {
-      printf("running host command: %s\n", hostcommand1);
+      log_printf(log_info, "running host command: %s", hostcommand1);
       system(hostcommand1);
     }
 
@@ -566,7 +971,7 @@ static int run_second_stage_script
   /* commit the change here */
   if (save_result)
     {
-      printf(" -> commit change to qemu image\n");
+      log_printf(log_info, "committing changes to qemu image");
       ret=forkexeclp("qemu-img", "qemu-img",
 		     "commit",
 		     cowdevpath,
@@ -575,15 +980,15 @@ static int run_second_stage_script
 
   /* after-run */
   loop_mount(workblockdevicepath, pc->buildplace);
-  printf(" -> running post-run process\n");
+  log_printf(log_info, "running post-run process");
   if(hostcommand2)
     {
-      printf("running host command: %s\n", hostcommand2);
+      log_printf(log_info, "running host command: %s", hostcommand2);
       system(hostcommand2);
     }
   loop_umount(pc->buildplace);
   rmdir(pc->buildplace);
-  printf(" -> clean up COW device files\n");
+  log_printf(log_info, "clean up COW device files");
   unlink(workblockdevicepath);
   unlink(cowdevpath);
   ret=0;
@@ -689,7 +1094,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
     {
       /* could not create the buildplace here. */
       ret=1;
-      perror("mkdir");
+      log_perror("mkdir");
       goto out;
     }
 
@@ -699,7 +1104,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
       goto out;
     }
 
-  printf(" -> Invoking debootstrap\n");
+  log_printf(log_info, "Invoking debootstrap");
   ret=forkexeclp("debootstrap", "debootstrap",
 		 "--arch",
 		 pc->arch,
@@ -710,8 +1115,7 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 		 NULL);
   if (ret)
     {
-      fprintf(stderr, "debootstrap failed with %i\n",
-	      ret);
+      log_printf(log_error, "debootstrap failed with %i", ret);
       goto umount_out;
     }
 
@@ -724,25 +1128,12 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
      we'll need think of other ways to work with them.
 
    */
-  printf(" -> Doing architecture-specific /dev population\n");
+  log_printf(log_info, "Doing architecture-specific /dev population");
 
   qemu_create_arch_devices(pc->buildplace, pc->arch);
 
   f = create_script(pc->buildplace, "pbuilder-run");
-  fprintf(f,
-	  "#!/bin/bash\n"
-	  "echo \n"
-	  "echo ' -> qemu-pbuilder first-stage' \n"
-	  "export PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'\n"
-	  "mount -n /proc /proc -t proc\n"
-	  "mount -n -o rw,remount / || mount -v -n -o rw,remount /dev/root \n"
-	  "cp /proc/mounts /etc/mtab\n"
-	  "export PBUILDER_MOUNTPOINT=/var/cache/pbuilder/pbuilder-mnt\n"
-	  "mkdir -p $PBUILDER_MOUNTPOINT\n"
-	  "mount -n -t ext3 /dev/%sb $PBUILDER_MOUNTPOINT \n"
-	  "$PBUILDER_MOUNTPOINT/pbuilder-run \n",
-	  qemu_arch_diskdevice(pc)
-	  );
+  write_first_stage(f, pc);
   fclose(f);
   f = NULL;
 
@@ -775,58 +1166,44 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 
   timestring=get_current_time_string();
 
-  f = create_script(pc->buildplace, "pbuilder-run");
-  fprintf(f,
-	  "#!/bin/bash\n"
-	  /* define function to terminate qemu */
-	  "function exit_from_qemu() {\n"
-	  "%s\n"
-	  "sync\n"
-	  "sync\n"
-	  "sleep 1s\n"		/* sleep before sending dying message */
-	  "echo ' -> qemu-pbuilder %s$1'\n"
-	  "sleep 1s\n"
-	  "halt -f -p\n"	/* just halt myself if possible */
-	  "}\n",
-	  pc->debug?"echo \"Debug shell\"; /bin/bash":"",
-	  qemu_keyword);
-
+  f = create_script(pc->buildplace, "input/pbuilder-run");
+  write_second_stage_header(f, pc->debug);
   fprintf(f,
 	  /* start of main code */
 	  "export RET=0\n"
 	  "echo \n"
-	  "echo ' -> qemu-pbuilder second-stage' \n"
-	  "echo '  -> setting time to %s' \n"
+	  "log.i 'qemu-pbuilder second-stage' \n"
+	  "log.i 'setting time to %s' \n"
 	  "date --set=\"%s\"\n"
-	  "echo '  -> Running debootstrap second-stage script' \n"
+	  "log.i 'Running debootstrap second-stage script' \n"
 	  "touch /etc/udev/disabled\n" // work-around for #520742
 	  "/debootstrap/debootstrap --second-stage || ( "
-	  "  echo dumping debootstrap log\n"
+	  "  log.i \"dumping debootstrap log\"\n"
 	  "  cat /debootstrap/debootstrap.log\n"
 	  "  exit_from_qemu\n"
 	  "\n )\n"
 	  "rm /etc/udev/disabled\n" // work-around for #520742
 	  "echo deb %s %s %s > /etc/apt/sources.list \n"
 	  "echo 'APT::Install-Recommends \"false\"; ' > /etc/apt/apt.conf.d/15pbuilder\n"
-	  //TODO: copy hook scripts
 	  "mount -n proc /proc -t proc\n"
 	  "mount -n sysfs /sys -t sysfs\n"
 	  "mkdir /dev/pts\n"
 	  "mount -n devpts /dev/pts -t devpts\n"
 	  "dhclient eth0\n"
-	  "$PBUILDER_MOUNTPOINT/run-copyfiles\n"
+	  "%s\n"
+	  "$BUILDDIR/input/run-copyfiles\n"
 	  "hostname pbuilder-$(cat /etc/hostname)\n"
 	  //TODO: installaptlines
 	  "echo '%s' > /etc/apt/sources.list.d/other.list\n"
-	  //TODO: run G hook
+	  EXECUTE_HOOKS("G")
 	  "apt-get update || exit_from_qemu 1\n"
 	  //TODO: "dpkg --purge $REMOVEPACKAGES\n"
 	  //recover aptcache
-	  "apt-get -y --force-yes -o DPkg::Options::=--force-confnew dist-upgrade || exit_from_qemu 1\n"
-	  "apt-get install --force-yes -y build-essential dpkg-dev apt aptitude pbuilder %s || exit_from_qemu 1\n"
+	  "apt-get -y %s -o DPkg::Options::=--force-confnew dist-upgrade || exit_from_qemu 1\n"
+	  "apt-get install %s -y build-essential dpkg-dev apt aptitude pbuilder %s || exit_from_qemu 1\n"
 	  //TODO: save aptcache
 	  //optionally autoclean aptcache
-	  //run E hook
+	  EXECUTE_HOOKS("E")
 	  //TODO: I can mount /var/cache/apt/archives from some scratch space to not need this:
 	  "apt-get clean || true\n"
 	  "exit_from_qemu $RET\n"
@@ -834,7 +1211,10 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 	  timestring,
 	  timestring,
 	  t=sanitize_mirror(pc->mirror), pc->distribution, pc->components,
+	  pc->hookdir && pc->hookdir[0] ? "mkdir -p \""CHROOT_HOOKDIR"\"" : "",
 	  pc->othermirror?pc->othermirror:"",
+	  pc->allow_untrusted?"--force-yes":"",
+	  pc->allow_untrusted?"--force-yes":"",
 	  pc->extrapackages?pc->extrapackages:"");
   fclose(f);
   f = NULL;
@@ -844,13 +1224,16 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
   /* TODO: can I do 'date --set' from output of 'LC_ALL=C date' */
 
   /* copy files script */
-  f = create_script(pc->buildplace, "run-copyfiles");
+  f = create_script(pc->buildplace, "input/run-copyfiles");
   copy_file_contents_through_temp(f, "/etc/hosts", pc->buildplace, "/etc");
   copy_file_contents_through_temp(f, "/etc/hostname", pc->buildplace, "/etc");
+  if (pc->hookdir != NULL && pc->hookdir[0])
+    {
+      copy_hookdir(pc->hookdir, pc->buildplace);
+    }
   fclose(f);
 
   /* do I not need to copy /etc/pbuilderrc, and ~/.pbuilderrc to inside chroot? */
-  /* TODO: hooks probably need copying here. */
   /* TODO: recover aptcache */
 
   loop_umount(pc->buildplace);
@@ -892,6 +1275,10 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
     asprintf(&debbuildopts_work, "%s -B", pc->debbuildopts);
     debbuildopts=debbuildopts_work;
     buildopt="--binary-arch";
+  } else if (pc->binary_indep) {
+    asprintf(&debbuildopts_work, "%s -A", pc->debbuildopts);
+    debbuildopts=debbuildopts_work;
+    buildopt="--binary-indep";
   } else {
     debbuildopts=pc->debbuildopts;
     buildopt="--binary-all";
@@ -900,14 +1287,23 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
   hoststr=copy_dscfile(dscfile, pc->buildplace);
 
   asprintf(&commandline,
-	   /* TODO: executehooks D: */
-	   "/usr/lib/pbuilder/pbuilder-satisfydepends --control $PBUILDER_MOUNTPOINT/*.dsc --internal-chrootexec 'chroot . ' %s \n"
-	   "apt-get install --force-yes -y %s\n"
-	   "cd $PBUILDER_MOUNTPOINT; /usr/bin/dpkg-source -x $(basename %s) \n"
-	   "echo ' -> Building the package'\n"
-	   /* TODO: executehooks A: */
-	   "cd $PBUILDER_MOUNTPOINT/*-*/; dpkg-buildpackage -us -uc %s\n",
+	   EXECUTE_HOOKS("D")
+	   "ALLOWUNTRUSTED=%s /usr/lib/pbuilder/pbuilder-satisfydepends --control $BUILDDIR/*.dsc --internal-chrootexec 'chroot . ' %s \n"
+	   "apt-get install %s -y %s\n"
+	   "cd $BUILDDIR; /usr/bin/dpkg-source -x $(basename %s) \n"
+	   "log.i 'Building the package'\n"
+	   EXECUTE_HOOKS("A")
+	   "if ! (\n"
+	   "    cd $BUILDDIR/*-*/\n"
+	   "    dpkg-buildpackage -us -uc %s\n"
+	   "); then\n"
+	   EXECUTE_HOOKS_INDENT("    ", "C")
+	   "    exit_from_qemu 1\n"
+	   "fi\n"
+	   EXECUTE_HOOKS("B"),
+	   pc->allow_untrusted?"yes":"no",
 	   buildopt,
+	   pc->allow_untrusted?"--force-yes":"",
 	   pc->extrapackages?pc->extrapackages:"",
 	   dscfile,
 	   debbuildopts);
@@ -916,8 +1312,23 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
      no other file will have _. */
 
   asprintf(&hoststr2,
-	   "cp -p \"%s\"/*_* \"%s\" 2>/dev/null || true",
-	   pc->buildplace, pc->buildresult);
+	   "BUILDPLACE='%s'\n"
+	   "BUILDRESULT='%s'\n"
+	   "BUILDRESULTUID=%d\n"
+	   "BUILDRESULTGID=%d\n"
+	   "if [ -d \"${BUILDRESULT}\" ]; then\n"
+	   "    chown \"${BUILDRESULTUID}:${BUILDRESULTGID}\" \"${BUILDPLACE}$BUILDDIR/\"*\n"
+	   "    chgrp \"${BUILDRESULTGID}\" \"${BUILDPLACE}$BUILDDIR/\"*\n"
+	   "    for FILE in \"${BUILDPLACE}$BUILDDIR\"/*; do\n"
+	   "        if [ -f \"${FILE}\" ]; then\n"
+	   "            cp -p \"${FILE}\" \"${BUILDRESULT}\" || true\n"
+	   "        fi\n"
+	   "    done\n"
+	   "else\n"
+	   "    log.e \"BUILDRESULT=[$BUILDRESULT] is not a directory.\"\n"
+	   "fi\n",
+	   pc->buildplace, pc->buildresult,
+	   pc->buildresultuid, pc->buildresultgid);
 
   ret=run_second_stage_script
     (0,
@@ -935,6 +1346,8 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
 int cpbuilder_login(const struct pbuilderconfig* pc)
 {
   return run_second_stage_script(pc->save_after_login,
+				 EXECUTE_HOOKS("H")
+				 EXECUTE_HOOKS("F")
 				 "bash",
 				 pc,
 				 NULL,
@@ -952,11 +1365,21 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
   char* runcommandline;
   int ret;
 
-  asprintf(&hostcommand, "cp %s %s/runscript\n", av[0], pc->buildplace);
+  asprintf(&hostcommand,
+	   "[ -d %s/input ] || mkdir %s/input\n || log.e \"Failed to create directory '%s/input'\"\n"
+	   "cp %s %s/input/runscript\n",
+	   pc->buildplace, pc->buildplace, pc->buildplace,
+	   av[0], pc->buildplace);
   /* TODO: add options too */
-  asprintf(&runcommandline, "sh $PBUILDER_MOUNTPOINT/runscript");
-  ret=run_second_stage_script(pc->save_after_login, runcommandline, pc,
-			      hostcommand, NULL);
+  asprintf(&runcommandline,
+	   EXECUTE_HOOKS("H")
+	   EXECUTE_HOOKS("F")
+	   "sh $BUILDDIR/input/runscript");
+  ret=run_second_stage_script(pc->save_after_login,
+			      runcommandline,
+			      pc,
+			      hostcommand,
+			      NULL);
   free(hostcommand);
   free(runcommandline);
   return ret;
@@ -969,26 +1392,78 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
  */
 int cpbuilder_update(const struct pbuilderconfig* pc)
 {
+  int ret;
+
+  ret = do_fsck(pc->basepath);
+  if (ret)
+    {
+      log_printf(log_error, "Failed fsck '%s'", pc->basepath);
+      return ret;
+    }
+
+  if (mkdir(pc->buildplace,0777))
+    {
+      log_printf(log_error,
+	  "Could not create directory '%s': %s",
+	  pc->buildplace,
+	  strerror(errno));
+      return 1;
+    }
+
+  ret = loop_mount(pc->basepath, pc->buildplace);
+  if (ret)
+    {
+      log_printf(log_error,
+	  "Could not mount '%s' on '%s'",
+	  pc->basepath,
+	  pc->buildplace);
+      return ret;
+    }
+  FILE *g = create_script(pc->buildplace, "pbuilder-run");
+  write_first_stage(g, pc);
+  fclose(g);
+  ret = loop_umount(pc->buildplace);
+  if (ret)
+    {
+      log_printf(log_error,
+	  "Could not unmount '%s'",
+	  pc->buildplace);
+      return ret;
+    }
+  ret = rmdir(pc->buildplace);
+  if (ret)
+    {
+      log_printf(log_error,
+	  "Could not rmdir '%s'",
+	  pc->buildplace);
+      return ret;
+    }
+
   /* TODO: --override-config support, --othermirror support etc.
      There is no way to change distribution in this code-path...
    */
   char *script;
-  if (0<asprintf(&script,
+  if (0>asprintf(&script,
+		 EXECUTE_HOOKS("H")
 		 //TODO: installaptlines if required.
 		 //TODO: "dpkg --purge $REMOVEPACKAGES\n"
 		 //TODO: add error code handling.
 		 "apt-get update -o Acquire::PDiffs=false\n"
-		 "apt-get -y --force-yes -o DPkg::Options::=--force-confnew dist-upgrade\n"
-		 "apt-get install --force-yes -y build-essential dpkg-dev apt aptitude pbuilder %s\n"
+		 "apt-get -y %s -o DPkg::Options::=--force-confnew dist-upgrade\n"
+		 "apt-get install %s -y build-essential dpkg-dev apt aptitude pbuilder %s\n"
 		 //TODO: optionally autoclean aptcache
-		 //run E hook
+		 EXECUTE_HOOKS("E")
+		 , pc->allow_untrusted?"--force-yes":""
+		 , pc->allow_untrusted?"--force-yes":""
 		 , pc->extrapackages?pc->extrapackages:""))
     {
-      fprintf(stderr, "qemubuilder: out of memory.\n");
+      log_printf(log_error, "qemubuilder: out of memory.");
       return 1;
     }
 
-  return run_second_stage_script(1, script, pc, NULL, NULL);
+  ret = run_second_stage_script(1, script, pc, NULL, NULL);
+  free(script);
+  return ret;
 }
 
 int cpbuilder_help(void)
